/* $Id: lambda.h,v 1.9 2001/07/20 21:42:03 reid Exp $ */

/* This file is the interface file for the module API and should be included
   by all modules. Note, however, that it contains things which are *not* part
   of the defined API (e.g. structure member names, etc.); don't troll this
   file for how-to information. The API manual is the definitive reference --
   if you can't figure it out from there, please file a bug report. */

#ifndef LAMBDA_H
#define LAMBDA_H

#include <config.h>
#include <stdlib.h> /* for size_t */


/*** Constants ***/

/* Global parameter defaults. */
#define DEFAULT_SAMPLE_RATE 44100
#define DEFAULT_SAMPLES_PER_TICK 256
#define DEFAULT_INPUT_MICROSECONDS 100000

/* Modules' magic number. */
#define MOD_MAGIC_1_0_0 0x6c010000

/* Maximum number of dependencies a port can have. */
#define MAX_PORT_DEPS 15


/*** Types ***/

typedef enum {
    OK, 
    ERR,     /* general error */
    ERR_EOF, /* end-of-file encountered */
} lerr;

/* Maximum size of a scalar object described by an ltyp structure. */
#define L_SIZEOF_MAX sizeof(double)

/* The following mess defines, with the help of autoconf, integers with
   specific sizes. C9x defines integers with specific types in <inttypes.h>,
   so trying those first is a possible FIXME [bug 39]. */

#if SIZEOF_INT == 2
    typedef int int16;
    typedef unsigned int uint16;
#elif SIZEOF_SHORT == 2
    typedef short int16;
    typedef unsigned short uint16;
#else
    "can't find a 16 bit integral type"
#endif

#if SIZEOF_INT == 4
    typedef int int32;
    typedef unsigned int uint32;
#elif SIZEOF_LONG == 4
    typedef long int32;
    typedef unsigned long uint32;
#else
    "can't find a 32 bit integral type"
#endif

#if SIZEOF_LONG_LONG == 8
    typedef long long int64;
    typedef unsigned long long uint64;
#else
    "can't find a 64 bit integral type"
#endif

/* Some more normal typedefs. */
typedef unsigned char byte;
typedef int bool;
typedef double sample;
typedef double freq;
typedef double dur;
typedef double level;

/* The list of "raw" ltyp type descriptors. Order here is very important; see
   the comment in lcore/types.c. */
typedef enum {
    LTYP_BOOL,
    LTYP_BYTE,
    LTYP_INT,
    LTYP_LONG,
    LTYP_INT16,
    LTYP_UINT16,
    LTYP_INT32,
    LTYP_UINT32,
    LTYP_INT64,
    LTYP_UINT64,
    LTYP_FLOAT,
    LTYP_DOUBLE,
    LTYP_SAMPLE,
    LTYP_FREQ,
    LTYP_DUR,
    LTYP_LEVEL
} raw_ltyp;

/* The list of "meta" ltyp descriptors -- whether an ltyp describes a scalar,
   array, etc. Order here is not important. */
typedef enum { 
    SCALAR,   /* n_elements ignored */
    ARRAY,    /* n_elements is size of array */
    ARRAY_REL /* n_elements * samples_per_tick() is size of array */
} meta_ltyp;

/* Da-da-da-dum, the ltyp type descriptor structure. */
typedef struct {
    meta_ltyp meta_type;
    raw_ltyp raw_type;
    unsigned n_elements;
} ltyp;

/* A couple of helper types for port and friends. */
enum port_direction {IN, OUT};
enum port_strictness {NONSTRICT = 0, STRICT};

/* Another central data structure, the port. This type is used for both input
   and output ports; the differences are as follows.

   Input ports: refcount is not used and should always be zero. If source is
   non-NULL, then the port is connected and source points to the source output
   port and contents points to source->contents. If source is NULL and
   contents is non-NULL, then the port is set to the value pointed to by
   contents. Otherwise, the port is unconnected and unset.

   Output ports: source is not used and should always be NULL. refcount is the
   number of connections, and contents points to the value of the port. */
struct port_struct {
    char * name;
    struct awid_struct * parent;
    enum port_direction direction;
    ltyp * type;
    enum port_strictness strict;
    char ** deps;
    unsigned refcount;
    struct port_struct * source;
    void * contents;
};
typedef struct port_struct port;

/* Port declaration. */
typedef struct {
    char * name;
    enum port_direction direction;
    meta_ltyp meta_type; /* this and the next two are identical to meta_ltyp */
    raw_ltyp raw_type;
    unsigned n_elements;
    enum port_strictness strict;
    char * depends[MAX_PORT_DEPS + 1];
} port_decl;

/* Port declaration list terminator. A port declaration with name == NULL
   terminates the list of declarations. The rest of this is arbitrary. */
#define END_PORT_DECLARATIONS { NULL, IN, SCALAR, 0, LTYP_INT, STRICT, {} }

/* The oscillator data structure. */
typedef struct {
    double phase;
    double duty;
} osc;


/*** Functions ***/

/* Functions defined by modules. */
lerr lm_go(void ** context, unsigned n_ports, port * ports[]);
lerr lm_pause(void * context);
lerr lm_stop(void * context);
lerr lm_tick(void * context);

/* Assertions. (l_assert.c) */
void _l_assert(char * file, int line);
#ifndef NDEBUG
#define l_ASSERT(expr) {if (!(expr)) _l_assert(__FILE__, __LINE__);}
#else
#define l_ASSERT(expr)
#endif

/* Type manipulation. (types.c) */
size_t l_sizeof(ltyp * t);
ltyp * ltyp_new(meta_ltyp meta_type, raw_ltyp raw_type, unsigned n_elements);
void ltyp_free(ltyp * t);
bool ltyp_eq(ltyp * t1, ltyp * t2);
char * ltyp_str(ltyp * t);
char * ltyp_tostr(ltyp * t, void * value);
lerr ltyp_fromstr(void * buf, ltyp * t, char * str);

/* Port operations. (ports.c) */
void l_send(port * p, void * buf);
void l_recv(port * p, void * buf);
port * port_addrof(char * name, unsigned n_ports, port * ports[]);
bool port_ready(port * p);
void _port_verify(port * p);
#ifndef NDEBUG
#define port_VERIFY(p) _port_verify(p)
#else
#define port_VERIFY(p)
#endif

/* Timing information. (timing.c) */
inline freq samples_per_second(void);
inline dur seconds_per_sample(void);
inline unsigned samples_per_tick(void);
inline unsigned samples_calc(void);
inline dur seconds_calc(void);

/* Oscillators. (osc.c) */
osc * osc_new(double phase, double duty);
void osc_free(osc * os);
void osc_sine(sample * buf, osc * os, freq f, unsigned count);
void osc_square(sample * buf, osc * os, freq f, unsigned count);
void osc_triangle(sample * buf, osc * os, freq f, unsigned count);
void osc_sawtooth(sample * buf, osc * os, freq f, unsigned count);

/* Memory management. (memory.c) */
void * l_malloc(size_t size);
void * l_realloc(void * p, size_t new_size);
void l_free(void * p);

/* Terminal output. (term_output.c) */
void pr(char * fmt, ...);
void error_pr(char * fmt, ...);
void debug_pr(char * fmt, ...);

/* Error handling. (errors.c) */
char * err_get(void);
char * err_copy(void);
void err_set(char * fmt, ...);
void err_prepend(char * fmt, ...);
void err_ignore_on(void);
void err_ignore_off(void);

#endif /* !LAMBDA_H */
