/* $Id: arena.c,v 1.3 2001/04/20 05:48:27 reid Exp $ */

/* This file implements the arena management stuff. */

#include "lambda.h"
#include "lcore.h"


/*** Globals ***/

arena g_arena = { NULL, STOP, 0 };


/*** Public Functions ***/

/* Cleanly make the arena empty, deleting all widgets. */
lerr arena_clear(void)
{
    awid * cur, * next;

    l_ASSERT(state() == STOP);

    cur = g_arena.first;
    while (cur != NULL) {
        next = cur->next;
        awid_free(cur);
        cur = next;
    }
    g_arena.first = NULL;

    g_arena.dirty = 0;
    return OK;
}

/* Load the module named mod_name and create an arena widget from it named
   awid_name; add that widget to the arena. */
lerr arena_load(char * mod_name, char * awid_name)
{
    awid * a;
    
    l_ASSERT(state() == STOP);

    a = awid_new(mod_name, awid_name);
    if (a == NULL)
        return ERR;

    a->next = g_arena.first;
    g_arena.first = a;

    g_arena.dirty = 1;
    return OK;
}

/* Remove the widget named awid_name and deallocate it. */
lerr arena_delete(char * awid_name)
{
    awid * cur, * prev;

    l_ASSERT(state() == STOP);

    if (g_arena.first != NULL) {
        if (!strcmp(g_arena.first->name, awid_name)) {
            cur = g_arena.first;
            g_arena.first = cur->next;
            awid_free(cur);
            g_arena.dirty = 1;
            return OK;
        } else {
            prev = g_arena.first;
            cur = prev->next;
            while (cur != NULL) {
                if (!strcmp(cur->name, awid_name)) {
                    prev->next = cur->next;
                    awid_free(cur);
                    g_arena.dirty = 1;
                    return OK;
                }
                prev = cur;
                cur = prev->next;
            }
        }
    }

    err_set("widget %s does not exist", awid_name);
    return ERR;
}

/* Connect the output port named src_port_name of the widget named
   src_awid_name to the input port named dst_port_name of the widget named
   dst_awid_name. On success, return OK; on error, set the error string and
   return ERR and do not change anything. */
lerr arena_connect(char * src_awid_name, char * src_port_name,
             char * dst_awid_name, char * dst_port_name)
{
    port * src_p, * dst_p;
    char * s;

    l_ASSERT(state() == STOP);

    if (!(src_p = awid_get_port(src_awid_name, src_port_name))
        || !(dst_p = awid_get_port(dst_awid_name, dst_port_name)))
        return ERR;
    port_VERIFY(src_p);
    port_VERIFY(dst_p);
    
    if (src_p->direction != OUT) {
        err_set("%s/%s: not an output port", src_p->parent->name,
                src_p->name);
        return ERR;
    }
    if (dst_p->direction != IN) {
        err_set("%s/%s: not an input port", dst_p->parent->name,
                dst_p->name);
        return ERR;
    }
    if (!ltyp_eq(src_p->type, dst_p->type)) {
        err_set("%s/%s -> %s/%s: types of ports do not match",
                src_p->parent->name, src_p->name, dst_p->parent->name,
                dst_p->name);
        return ERR;
    }
    if (dst_p->source != NULL) {
        err_set("%s/%s: already connected to by %s/%s", dst_awid_name,
                dst_port_name, dst_p->source->name,
                dst_p->source->parent->name);
        return ERR;
    }
    if (dst_p->contents != NULL) {
        s = ltyp_tostr(dst_p->type, dst_p->contents);
        err_set("%s/%s: already set to %s\n",
                dst_awid_name, dst_port_name, s);
        l_free(s);
        return ERR;
    }

    src_p->refcount++;
    if (src_p->refcount == 1)
        src_p->contents = l_malloc(l_sizeof(src_p->type));
    dst_p->source = src_p;
    dst_p->contents = src_p->contents;

    port_VERIFY(src_p);
    port_VERIFY(dst_p);
    g_arena.dirty = 1;
    return OK;
}

/* Delete the connection from the port named src_port_name on the widget named
   src_awid_name to the port named dst_port_name on the widget named
   dst_awid_name. */
lerr arena_disconnect(char * src_awid_name, char * src_port_name,
                char * dst_awid_name, char * dst_port_name)
{
    port * src_p;
    port * dst_p;

    l_ASSERT(state() == STOP);

    if (!(src_p = awid_get_port(src_awid_name, src_port_name))
        || !(dst_p = awid_get_port(dst_awid_name, dst_port_name)))
        return ERR;
    port_VERIFY(src_p);
    port_VERIFY(dst_p);
    
    if (src_p->direction != OUT) {
        err_set("%s/%s: not an output port", src_p->parent->name,
                src_p->name);
        return ERR;
    }
    if (dst_p->direction != IN) {
        err_set("%s/%s: not an input port", dst_p->parent->name,
                dst_p->name);
        return ERR;
    }
    if (dst_p->source != src_p) {
        err_set("%s/%s -> %s/%s: not connected", src_awid_name,
                src_port_name, dst_awid_name, dst_port_name);
        return ERR;
    }

    l_ASSERT(src_p->refcount > 0);
    src_p->refcount--;
    if (src_p->refcount == 0) {
        l_free(src_p->contents);
        src_p->contents = NULL;
    }
    dst_p->source = NULL;
    dst_p->contents = NULL;

    port_VERIFY(src_p);
    port_VERIFY(dst_p);
    g_arena.dirty = 1;
    return OK;
}

/* Set the input port named dst_port_name on the widget named dst_awid_name to
   a value whose string representation is s. */
lerr arena_set(char * dst_awid_name, char * dst_port_name, char * s)
{
    port * dst_p;
    bool first;
    
    if (!(dst_p = awid_get_port(dst_awid_name, dst_port_name)))
        return ERR;
    port_VERIFY(dst_p);

    if (dst_p->direction != IN) {
        err_set("%s/%s: not an input port", dst_awid_name, dst_port_name);
        return ERR;
    }
    if (dst_p->source != NULL) {
        err_set("%s/%s -> %s/%s: already connected", dst_awid_name,
                dst_port_name, dst_p->source->name,
                dst_p->source->parent->name);
        return ERR;
    }
    
    /* dst_p->contents is altered in-place to avoid calling l_malloc() if it's
       being reset, because it is preferred not to call l_malloc() when not in
       the stop state. Hence the messiness backing out the memory allocation
       if s is not interpretable. */
    if (dst_p->contents == NULL) {
        if (state() != STOP) {
            err_set("can set a previously unset port only in stop state");
            return ERR;
        }
        dst_p->contents = l_malloc(l_sizeof(dst_p->type));
        first = 1;
    } else
        first = 0;
    if (ltyp_fromstr(dst_p->contents, dst_p->type, s) != OK) {
        err_prepend("cannot interpret `%s': ", s);
        if (first) {
            l_free(dst_p->contents);
            dst_p->contents = NULL;
        }
        return ERR;
    }

    port_VERIFY(dst_p);
    if (first)
        g_arena.dirty = 1;
    return OK;
}

/* Unset the port named dst_port_name on the widget named dst_awid_name. */
lerr arena_unset(char * dst_awid_name, char * dst_port_name)
{
    port * dst_p;
    
    l_ASSERT(state() == STOP);

    if (!(dst_p = awid_get_port(dst_awid_name, dst_port_name)))
        return ERR;
    port_VERIFY(dst_p);

    if (dst_p->direction != IN) {
        err_set("%s/%s: not an input port", dst_awid_name, dst_port_name);
        return ERR;
    }
    if (!(dst_p->source == NULL && dst_p->contents != NULL)) {
        err_set("%s/%s: not set", dst_awid_name, dst_port_name);
        return ERR;
    }

    l_free(dst_p->contents);
    dst_p->contents = NULL;

    port_VERIFY(dst_p);
    g_arena.dirty = 1;
    return OK;
}

/* Verify that the arena is ready to enter the go state--check that each
   module's port dependencies are satisfied. If so, return OK and clear the
   arena dirty bit; otherwise, return ERR and set the error string
   appropriately. */
lerr arena_ready(void)
{
    awid * a;

    a = g_arena.first;
    while (a != NULL) {
        if (awid_check_deps(a) != OK) {
            l_ASSERT(g_arena.dirty);
            err_prepend("%s: ", a->name);
            return ERR;
        }
        a = a->next;
    }

    g_arena.dirty = 0;
    return OK;
}

/* Pretty-print the arena. */
lerr arena_print(void)
{
    awid * cur;

    pr("Arena is %s.\n", g_arena.dirty ? "dirty" : "clean");

    cur = g_arena.first;
    while (cur != NULL) {
        awid_pr(cur);
        cur = cur->next;
    }

    return OK;
}
