/* $Id: awidgets.c,v 1.5 2001/07/20 21:42:03 reid Exp $ */

/* This file implements the widget manipulation functions. */

#include "lambda.h"
#include "lcore.h"

/*** Globals ***/

extern arena g_arena;


/*** Public functions ***/

/* Build an arena widget, named awid_name, as an instance of the module named
   mod_name. Sets up all fields but does not link the new awid into the
   arena's linked list -- the next field is set to NULL. On success, return a
   pointer to the node; on error, return NULL and set the error string. */
awid * awid_new(char * mod_name, char * awid_name)
{
    awid * a;
    module * m;
    int i, j;

    if (awid_addrof(awid_name)) {
        err_set("widget name %s is in use", awid_name);
        return NULL;
    }

    m = load_module(mod_name);
    if (m == NULL)
        return NULL;

    a = l_malloc(sizeof(awid));
    a->mod = m;
    a->name = l_malloc(strlen(awid_name) + 1);
    strcpy(a->name, awid_name);
    a->next = NULL;

    a->n_ports = 0;
    a->ports = NULL;
    for (i = 0; a->mod->decls[i].name != NULL; i++) {
        /* Sanity check, in case the module forgot to terminate the
           declaration list or something else is broken. FIXME [bug 24]: this
           should go into a decls_verify() function. */
        if (!(a->mod->decls[i].direction == IN
              || a->mod->decls[i].direction == OUT)
            || !(a->mod->decls[i].meta_type == SCALAR
                 || a->mod->decls[i].meta_type == ARRAY
                 || a->mod->decls[i].meta_type == ARRAY_REL)
            || strlen(a->mod->decls[i].name) == 0) {
            err_set("module %s is broken: object `ports[%d]' is invalid",
                    a->mod->name, i);
            awid_free(a);
            return NULL;
        }

        /* Verify that the port name is unique. */
        for (j = 0; j < a->n_ports; j++)
            if (!strcmp(a->mod->decls[i].name, a->ports[j]->name)) {
                err_set("module %s is broken: duplicate port name %s",
                        a->mod->name, a->mod->decls[i].name);
                awid_free(a);
                return NULL;
            }

        a->n_ports++;
        a->ports = l_realloc(a->ports, a->n_ports * sizeof(port *));
        a->ports[a->n_ports - 1] = port_new(a, &(a->mod->decls[i]));
    }

    /* Verify that each port has a valid number of dependencies and that no
       port depends on a nonexistent port. */
    for (i = 0; i < a->n_ports; i++) {
        for (j = 0; a->ports[i]->deps[j] != NULL; j++) {
            if (j == MAX_PORT_DEPS) {
                err_set("module %s is broken: too many port dependencies",
                        a->mod->name);
                awid_free(a);
                return NULL;
            }
            if (!port_addrof(a->ports[i]->deps[j], a->n_ports, a->ports)) {
                err_set("module %s is broken: port %s depends on nonexistent"
                        " port %s", a->mod->name, a->ports[i]->name,
                        a->ports[i]->deps[j]);
                awid_free(a);
                return NULL;
            }
        }
    }

    return a;
}

/* Free all storage associated with a. Postcondition: a points to garbage. */
void awid_free(awid * a)
{
    int i;

    /* FIXME [bug 25]: need to break all connections involving a. */

    unload_module(a->mod);
    l_free(a->name);
    for (i = 0; i < a->n_ports; i++)
        port_free(a->ports[i]);
    l_free(a->ports);
    l_free(a);
}

/* Return the address of the widget in the arena named name, or NULL if no
   widget matches. */
awid * awid_addrof(char * name)
{
    awid * cur;

    cur = g_arena.first;
    while (cur != NULL) {
        if (!strcmp(cur->name, name))
            return cur;
        cur = cur->next;
    }

    return NULL;
}

/* If the widget named awid_name has an input port named port_name, return a
   pointer to it; otherwise, set the error string and and return NULL. */
port * awid_get_port(char * awid_name, char * port_name)
{
    awid * a;
    port * p;

    a = awid_addrof(awid_name);
	if (a == NULL) {
        err_set("widget %s does not exist", awid_name);
        return NULL;
    }
    p = port_addrof(port_name, a->n_ports, a->ports);
    if (p == NULL) {
        err_set("widget %s has no input port %s", awid_name, port_name);
        return NULL;
    }
    
    return p;
}

/* Verify that all of an arena widget a's port dependencies are satisfied. If
   so, return OK; if not, return ERR and set the error string. */
lerr awid_check_deps(awid * a)
{
    port * target;
    int i, j;

    for (i = 0; i < a->n_ports; i++) {
        port_VERIFY(a->ports[i]);
        if (port_ready(a->ports[i]) || a->ports[i]->strict) {
            for (j = 0; a->ports[i]->deps[j] != NULL; j++) {
                target = port_addrof(a->ports[i]->deps[j], a->n_ports,
                                     a->ports);
                l_ASSERT(target != NULL);
                if (!port_ready(target)) {
                    if (port_ready(a->ports[i])) {
                        err_set("port %s depends on %s, not satisfied",
                                a->ports[i]->name, target->name);
                        return ERR;
                    } else
                        break;
                }
            }

            if (a->ports[i]->deps[j] == NULL
                && a->ports[i]->strict && !port_ready(a->ports[i])) {
                err_set("port %s is strict and not ready", a->ports[i]->name);
                return ERR;
            }
        }
    }

    return OK;
}

/* Pretty-print an arena widget a. */
void awid_pr(awid * a)
{
    int i, j;
    char * s;

    pr("Widget %s (%s %s).\n", a->name, a->mod->name,
       a->mod->version);
    for (i = 0; i < a->n_ports; i++) {
        s = ltyp_str(a->ports[i]->type);
        pr("    %s: %s %s", a->ports[i]->name,
           (a->ports[i]->direction == IN) ? "in" : "out", s);
        l_free(s);
        if (a->ports[i]->direction == IN) {
            if (a->ports[i]->source != NULL) {
                pr(" <- %s/%s", a->ports[i]->source->parent->name,
                   a->ports[i]->source->name);
            } else if(a->ports[i]->contents != NULL) {
                s = ltyp_tostr(a->ports[i]->type, a->ports[i]->contents);
                pr(" <- %s", s);
                l_free(s);
            }
        } else {
            /* FIXME [bug 26]: also print connections for output ports */
        }
        pr(" <%s", a->ports[i]->strict ? "strict" : "nonstrict" );
        for (j = 0; a->ports[i]->deps[j] != NULL; j++) {
            pr(" %s", a->ports[i]->deps[j]);
        }
        pr(">", s);
        pr("\n");
    }
}
