/* $Id: lcore.h,v 1.8 2001/07/23 05:36:08 reid Exp $ */

/* This file defines all interfaces exported by lcore source files for use by
   other parts of lcore. */

#ifndef LCORE_H
#define LCORE_H

#include "lambda.h"

#include <stdio.h> /* for FILE structure */


/*** Crud ***/

/* FIXME [bug 27]: for some reason we're not finding the prototypes for these
   two functions. Look into it... */
#include <stdarg.h>
int asprintf(char ** ptr, const char * fmt, ...);
int vasprintf(char ** ptr, const char * fmt, va_list ap);


/*** Constants ***/

/* When speed is limited (i.e. when running in "real-time mode"), leave this
   much buffer time -- i.e., if a tick finishes calculation early, sleep until
   this much time is left. This should be as small as possible.

   FIXME [bug 28]: make this a command-line option. */
#define SLUSH_USECS 1000

/* Module search path environment variable name, default value, and delimiter
   character. FIXME: make autoconf-ed. */
#define MOD_PATH_VAR     "LAMBDA_PATH"
#define MOD_PATH_DEFAULT "/usr/local/lib/lambda/modules"
#define MOD_PATH_DELIM   ":"

/* Suffix for module files. */
#define MOD_SUFFIX ".lm"

/* If usleep()ing by less than this, busy-wait instead of actually sleeping. */
#define USLEEP_BUSYWAIT_USECS 10

/* The maximum length of an error string, including terminator byte. */
#define ERR_CHARS_MAX 4096


/*** Types ***/

/* Module. */
typedef struct {
    void * handle;
    char * name;
    char * version;
    char * desc;
    port_decl * decls;
    lerr (*go)(void **, unsigned, port **);
    lerr (*pause)(void *);
    lerr (*stop)(void *);
    lerr (*tick)(void *);
} module;

/* Arena widget. */
struct awid_struct {
    char * name;
    module * mod;
    unsigned n_ports;
    port ** ports;
    void * context;
    struct awid_struct * next;
};
typedef struct awid_struct awid;

/* Set of possible arena states. */
typedef enum { STOP, PAUSE, GO } state_t;

/* Arena. */
typedef struct {
    awid * first;
    state_t state;
    bool dirty;
} arena;


/*** Functions ***/

/* cli.c */
lerr cli_process_line(FILE * fp);
void cli_prompt(void);

/* modules.c */
void modules_init(void);
module * load_module(char * name);
void unload_module(module * mod);

/* arena.c */
lerr arena_clear(void);
lerr arena_load(char * mod_name, char * awid_name);
lerr arena_delete(char * awid_name);
lerr arena_connect(char * src_awid_name, char * src_port_name,
                   char * dst_awid_name, char * dst_port_name);
lerr arena_disconnect(char * src_awid_name, char * src_port_name,
                      char * dst_awid_name, char * dst_port_name);
lerr arena_set(char * dst_awid_name, char * dst_port_name, char * str);
lerr arena_unset(char * dst_awid_name, char * dst_port_name);
lerr arena_print(void);
lerr arena_ready(void);

/* awidgets.c */
awid * awid_new(char * mod_name, char * awid_name);
void awid_free(awid * a);
awid * awid_addrof(char * name);
port * awid_get_port(char * awid_name, char * port_name);
void awid_pr(awid * a);
lerr awid_check_deps(awid * a);

/* states.c */
lerr enter_go(void);

lerr enter_pause(void);
lerr enter_stop(awid * limit);
lerr do_tick(void);
char * state_str(void);
state_t state(void);

/* ports.c */
port * port_new(struct awid_struct * parent, port_decl * decl);
void port_free(port * p);

/* timing.c */
uint64 usecs_wall(void);
void l_usleep(uint64 usecs);
void timing_init(freq sr, unsigned tl, bool limit);
void timing_pausego(void);
void timing_stopgo(void);
void timing_tick_start(void);
void timing_tick_end(void);
double timing_utilization(void);
void timing_stats_pr(void);

/* util.c */
bool l_feof(FILE * fp);
int min(int a, int b);

#endif /* !LCORE_H */
