/* $Id: main.c,v 1.9 2001/07/23 05:36:08 reid Exp $ */

/* This file contains the startup/shutdown stuff and the main loop. */

#include "lambda.h"
#include "lcore.h"

#include <getopt.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>


/*** Globals ***/

unsigned input_usecs; /* minimum time to wait before checking for input */
extern bool g_print_debug; /* from term_output.c */


/*** Private prototypes ***/

void setup(int argc, char ** argv);
int shutdown(void);


/*** Public functions ***/

int main(int argc, char ** argv)
{
    lerr result;
    uint64 input_last;

    pr("lambda version %s starting, please mind the rotor wash.\n",
       VERSION);

    setup(argc, argv);
    
    l_ASSERT(state() == STOP);
    result = OK;
    cli_prompt();
    while (result != ERR_EOF) {
        if (state() == GO) {
            timing_tick_start();
            result = do_tick();
            if (result != OK) {
                pr("\n? %s\n", err_get());
                result = enter_stop(NULL);
                if (result != OK)
                    pr("? %s\n", err_get());
                cli_prompt();
            }
            if (usecs_wall() > input_last + input_usecs) {
                input_last = usecs_wall();
                if (!l_feof(stdin)) {
                    result = cli_process_line(stdin);
                    if (result == ERR)
                        pr("? %s\n", err_get());
                    cli_prompt();
                }
            }
            timing_tick_end();
        } else {
            input_last = usecs_wall();
            result = cli_process_line(stdin);
            if (result == ERR)
                pr("? %s\n", err_get());
            cli_prompt();
        }
    }

    return shutdown();
}


/*** Private functions ***/

/* Set up the program, including command line option parsing. */
void setup(int argc, char ** argv)
{
    unsigned seed;
    freq sample_rate;
    unsigned tick_len;
    bool limit_realtime;
    int opt, optidx;
    const struct option options[] = { 
        {"seed", required_argument, NULL, 's'},
        {"sample-rate", required_argument, NULL, 'r'},
        {"samples-per-tick", required_argument, NULL, 't'},
        {"input-microseconds", required_argument, NULL, 'i'},
        {"debug", no_argument, NULL, 'd'},
        {"no-speed-limit", no_argument, NULL, 'l'},
        {NULL, 0, 0, 0 } };
            
    /* Set globals to their defaults. */
    sample_rate = DEFAULT_SAMPLE_RATE;
    tick_len = DEFAULT_SAMPLES_PER_TICK;
    limit_realtime = 1;
    input_usecs = DEFAULT_INPUT_MICROSECONDS;
    g_print_debug = 0;
    seed = time(NULL);

    /* Process command line args. */
    /* FIXME [bug 29]: check validity of input. */
    while ((opt = getopt_long(argc, argv, "s:r:t:i:dl", options, &optidx))
           != -1) {
        switch (opt) {
        case 's':
            seed = (unsigned int) strtoul(optarg, NULL, 10);
            break;
        case 'r':
            sample_rate = (freq) strtod(optarg, NULL);
            break;
        case 't':
            tick_len = (unsigned) strtoul(optarg, NULL, 10);
            break;
        case 'i':
            input_usecs = (unsigned) strtoul(optarg, NULL, 10);
            break;
        case 'd':
            g_print_debug = 1;
            break;
        case 'l':
            limit_realtime = 0;
            break;
        default:
            l_ASSERT(0);
            break;
        }
    }

    /* Initialize subsystems. */
    srand(seed);
    timing_init(sample_rate, tick_len, limit_realtime);
    modules_init();

    /* Tell the world. */
    debug_pr("debugging output: enabled\n");
#ifndef NDEBUG
    debug_pr("assertions:       enabled\n");
#else
    debug_pr("assertions:       disabled\n");
#endif
    debug_pr("\n");
    debug_pr("random seed: %u\n", seed);
    debug_pr("sample rate: %f\n", sample_rate);
    debug_pr("samples per tick: %u\n", tick_len);
    debug_pr("input microseconds: %u\n", input_usecs);
    debug_pr("speed limiting: %s\n", limit_realtime ? "yes" : "no");
}

/* Clean up. */
int shutdown(void)
{
    pr("\nGoodbye.\n");

    return 0;
}
