/* $Id: modules.c,v 1.8 2001/07/23 05:36:08 reid Exp $ */

/* This file contains the module loading and unloading functions. */

#include "lambda.h"
#include "lcore.h"

#include <dlfcn.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>


/*** Globals ***/

/* Null-terminated list of directories in the module search path. */
char ** mod_path;


/*** Public functions ***/

/* Initialize the module search path. */
void modules_init(void)
{
     char * dir;
     char * s; /* Note that elements of mod_path will point into this string
                  for the duration of the program; don't free it. */
     int dir_count;

     if (getenv(MOD_PATH_VAR))
          s = strdup(getenv(MOD_PATH_VAR));
     else
          s = strdup(MOD_PATH_DEFAULT);

     if (s == NULL)
          /* memabort("strdup"); FIXME [bug 30] */
          abort();

     mod_path = NULL;
     dir_count = 0;
     dir = strtok(s, MOD_PATH_DELIM);
     while (dir != NULL) {
          dir_count++;
          mod_path = l_realloc(mod_path, dir_count * sizeof(char *));
          mod_path[dir_count - 1] = dir;
          dir = strtok(NULL, MOD_PATH_DELIM);
     }
     mod_path = l_realloc(mod_path, (dir_count + 1) * sizeof(char *));
     mod_path[dir_count] = NULL;
}

/* Find and load the module named name and return a pointer to it. Return NULL
   and set the error string if an error occurred. */
module * load_module(char * name)
{
    module * m;
    char * filename, * s;
    int i, result;
    unsigned * magic;
    struct stat ss;

    /* Find it. */
    for (i = 0; mod_path[i] != NULL; i++) {
         asprintf(&filename, "%s/%s%s", mod_path[i], name, MOD_SUFFIX);
         if (filename == NULL)
              /* memabort("asprintf"); FIXME [bug 30] */
              abort();
         result = stat(filename, &ss);
         if (result == 0 || (result == -1 && errno != ENOENT))
              break;
         l_free(filename);
         filename = NULL;
    }
    if (filename == NULL) {
         err_set("can't find module %s", name);
         goto error;
    }

    /* Load it. */
    m = l_malloc(sizeof(module));
    m->handle = dlopen(filename, RTLD_NOW);
    l_free(filename);
    if (m->handle == NULL) {
        err_set("can't load module %s: %s", name, dlerror());
        goto error;
    }
    magic = dlsym(m->handle, "magic");
    if ((s = dlerror()) != NULL) {
        err_set("module %s is broken: missing object `magic'", name);
        goto error;
    }
    if (*magic != MOD_MAGIC_1_0_0) {
        err_set("module %s is broken: bad magic number %x (expected %x)",
                  name, *magic, MOD_MAGIC_1_0_0);
        goto error;
    }
    m->name = dlsym(m->handle, "name");
    if ((s = dlerror()) != NULL) {
        err_set("module %s is broken: missing object `name'", name);
        goto error;
    }
    m->version = dlsym(m->handle, "version");
    if ((s = dlerror()) != NULL) {
        err_set("module %s is broken: missing object `version'", name);
        goto error;
    }
    m->desc = dlsym(m->handle, "description");
    if ((s = dlerror()) != NULL) {
        err_set("module %s is broken: missing object `description'", name);
        goto error;
    }
    m->decls = dlsym(m->handle, "port_decls");
    if ((s = dlerror()) != NULL) {
        err_set("module %s is broken: missing object `port_decls'", name);
        goto error;
    }
    m->go = dlsym(m->handle, "lm_go");
    if ((s = dlerror()) != NULL) {
        err_set("module %s is broken: missing object `lm_go'", name);
        goto error;
    }
    m->pause = dlsym(m->handle, "lm_pause");
    if ((s = dlerror()) != NULL) {
        err_set("module %s is broken: missing object `lm_pause'", name);
        goto error;
    }
    m->stop = dlsym(m->handle, "lm_stop");
    if ((s = dlerror()) != NULL) {
        err_set("module %s is broken: missing object `lm_stop'", name);
        goto error;
    }
    m->tick = dlsym(m->handle, "lm_tick");
    if ((s = dlerror()) != NULL) {
        err_set("module %s is broken: missing object `lm_tick'", name);
        goto error;
    }

    return m;

error:
    l_free(m);
    return NULL;
}

/* Unload the module mod. Postcondition: mod points to garbage.*/
void unload_module(module * mod)
{
    dlclose(mod->handle);
    l_free(mod);
}
