/* $Id: ports.c,v 1.5 2001/07/20 21:42:03 reid Exp $ */

/* This file implements the data movement functions. */

#include "lambda.h"
#include "lcore.h"

#include <string.h> /* for memcpy() */


/*** Public functions ***/

/* Send the data stored in buf on port p. If the type of the data in buf is
   not described by p->type, the results are undefined. */
void l_send(port * p, void * buf)
{
    port_VERIFY(p);
    l_ASSERT(p->direction == OUT);

    memcpy(p->contents, buf, l_sizeof(p->type));
}

/* Copy the data waiting on port p to local buffer buf. It is the caller's
   responsibility to ensure that buf is large enough; if it is not the results
   are undefined. */
void l_recv(port * p, void * buf)
{
    port_VERIFY(p);
    l_ASSERT(p->direction == IN);

    memcpy(buf, p->contents, l_sizeof(p->type));
}

/* If there is a port named name in ports (whose size is n_ports), return its
   address; otherwise, return NULL. */
port * port_addrof(char * name, unsigned n_ports, port * ports[])
{
    unsigned i;

    for (i = 0; i < n_ports; i++)
        if (!strcmp(name, ports[i]->name)) {
            port_VERIFY(ports[i]);
            return ports[i];
        }

    return NULL;
}

/* Allocate and initialize a new port according to its declaration decl and
   parent node parent; return a pointer to it. Cannot fail. */
port * port_new(struct awid_struct * parent, port_decl * decl)
{
    port * p;

    p = l_malloc(sizeof(port));
    p->name = decl->name;
    p->parent = parent;
    p->direction = decl->direction;
    p->type = ltyp_new(decl->meta_type, decl->raw_type, decl->n_elements);
    p->strict = decl->strict;
    p->deps = decl->depends;
    p->refcount = 0;
    p->source = NULL;
    p->contents = NULL;

    port_VERIFY(p);
    return p;
}

/* Free a port. Postcondition: p points to garbage. */
void port_free(port * p)
{
    port_VERIFY(p);
    ltyp_free(p->type);
    l_free(p);
}

/* Return 1 if p is ready for data transfer (i.e. connected or set); 0
   otherwise. */
bool port_ready(port * p)
{
    port_VERIFY(p);

    if (p->direction == IN)
        return (bool) p->contents;
    else
        return (bool) p->refcount;
}

/* Check that a port is internally consistent. If not, abort. Only works if
   assertions are enabled. */
void _port_verify(port * p)
{
    l_ASSERT(p != NULL);
    l_ASSERT(p->strict == STRICT || p->strict == NONSTRICT);
    if (p->direction == IN) {
        l_ASSERT(p->refcount == 0);
        if (p->source != NULL) {
            l_ASSERT(p->contents != NULL);
            l_ASSERT(p->source->contents != NULL);
            l_ASSERT(ltyp_eq(p->type, p->source->type));
        }
    } else if (p->direction == OUT) {
        l_ASSERT(p->source == NULL);
        if (p->refcount > 0)
            l_ASSERT(p->contents != NULL);
    } else
        l_ASSERT(0);
}
