/* $Id: states.c,v 1.3 2001/04/20 05:48:27 reid Exp $ */

/* This file contains state management stuff and do_tick(). */

#include "lambda.h"
#include "lcore.h"


/*** Globals ***/

extern arena g_arena;


/*** Public functions ***/

/* Execute one tick's worth of calculations. If an error occurs, set the error
   string and return ERR (in this case, the arena is in an inconsistent state
   and should be stopped immediately); return OK on success. This function is
   only appropriate in the go state. */
lerr do_tick(void)
{
    awid * cur = g_arena.first;

    l_ASSERT(state() == GO);

    while (cur != NULL) {
        if ((* cur->mod->tick)(cur->context) != OK) {
            err_prepend("%s: ", cur->name);
            return ERR;
        }
        cur = cur->next;
    }

    return OK;
}

/* Return the current state. */
state_t state(void)
{
    return g_arena.state;
} 

/* Return a string describing the current state. */
char * state_str(void)
{
    switch (state()) {
    case GO:
        return "go";
        break;
    case PAUSE:
        return "pause";
        break;
    case STOP:
        return "stop";
        break;
    default:
        l_ASSERT(0);
    }

    l_ASSERT(0);
    return NULL;
}

/* Attempt to enter the go state. On success, return OK; on failure, set the
   error string and return ERR. */
lerr enter_go(void)
{
    awid * cur;
    char * s;

    if (g_arena.dirty) {
        err_set("arena is dirty");
        return ERR;
    }

    if (state() == STOP) {
        cur = g_arena.first;
        while (cur != NULL) {
            if((* cur->mod->go)(&(cur->context), cur->n_ports, cur->ports)
               != OK) {
                err_prepend("go failed: %s: ", cur->name);
                /* set all widgets already in go state back to stop, saving
                   error string in case going back to stop fails too. */
                s = err_copy();
                if (enter_stop(cur) != OK)
                    err_prepend("%s\n", s);
                l_free(s);
                return ERR;
            }
            cur = cur->next;
        }
        timing_stopgo();
    } else {
        l_ASSERT(state() == PAUSE);
        timing_pausego();
    }
    
    g_arena.state = GO;
    return OK;
}

/* Enter the pause state. On success, return OK; on error, set the error
   string and return ERR, but still go to the pause state. Try to set all
   widgets to pause even if some fail. */
lerr enter_pause(void)
{
    awid * cur;
    char ** err_list = NULL;
    int n_errs = 0;

    l_ASSERT(state() == GO);

    cur = g_arena.first;
    while (cur != NULL) {
        if((* cur->mod->pause)(cur->context) != OK) {
            n_errs++;
            err_prepend("%s: ", cur->name);
            err_list = l_realloc(err_list, n_errs * sizeof(char *));
            err_list[n_errs - 1] = err_copy();
        }
        cur = cur->next;
    }

    g_arena.state = PAUSE;

    if (n_errs == 0)
        return OK;
    else {
        err_set("");
        while (n_errs > 0) {
            err_prepend("\n  %s", err_list[n_errs - 1]);
            l_free(err_list[n_errs - 1]);
            n_errs--;
        }
        err_prepend("pause failed:");
        l_free(err_list);
        return ERR;
    }
}

/* Enter the stop state. If limit != NULL, only set widgets before limit to
   stop. Try to set all appropriate widgets to stop even if some fail. On
   success, return OK; on failure, set the error string and return ERR, but
   still enter the stop state. */
lerr enter_stop(awid * limit)
{
    awid * cur;
    char ** err_list = NULL;
    int n_errs = 0;

    cur = g_arena.first;
    while (cur != limit) {
        l_ASSERT(cur != NULL);
        if((* cur->mod->stop)(cur->context) != OK) {
            n_errs++;
            err_prepend("%s: ", cur->name);
            err_list = l_realloc(err_list, n_errs * sizeof(char *));
            err_list[n_errs - 1] = err_copy();
        }
        cur = cur->next;
    }

    g_arena.state = STOP;
    
    if (n_errs == 0)
        return OK;
    else {
        err_set("");
        while (n_errs > 0) {
            err_prepend("\n  %s", err_list[n_errs - 1]);
            l_free(err_list[n_errs - 1]);
            n_errs--;
        }
        err_prepend("stop failed:");
        l_free(err_list);
        return ERR;
    }
}

