/* $Id: timing.c,v 1.8 2001/07/20 21:42:03 reid Exp $ */

/* This file contains the time management functions. */

#include "lambda.h"
#include "lcore.h"

#include <sys/time.h>
#include <time.h>


/*** Constants ***/

#define ONE_MILLION 1000000


/*** Globals ***/

/* Sample rate in hertz. */
freq sample_rate;

/* Number of samples per tick. */
unsigned tick_len;

/* If true, don't calculate faster than real time. */
bool limit_realtime;

/* Number of samples calculated since the last stop->go transition, up to the
   end of the last tick. */
uint64 samples_elapsed;

/* Number of samples calculated since the last transition into the go state,
   up to the end of the last tick. */
uint64 go_samples;

/* Time spent calculating since the last go transition into the go state, up
   to the end of the last tick. */
uint64 go_usecs;

/* Wall time of the last transition into the go state. */
uint64 go_entry_wall;

/* Wall time of the beginning of the last tick. */
uint64 tick_start_wall;


/*** Public functions ***/

/* Return the number of microseconds since the epoch (a specific time in the
   past which is unchanged for the life of the program). This unsigned 64-bit
   quantity rolls over after about 585,000 years. Whether or not that is a bug
   is an exercise for the reader; this program has no provisions for 64-bit
   variables rolling over. */
uint64 usecs_wall(void)
{
    struct timeval tv;

    gettimeofday(&tv, NULL); /* this call cannot fail */
    return tv.tv_sec * (uint64)ONE_MILLION + tv.tv_usec;
}

/* Try to sleep for usecs microseconds. The time slept may end up being longer
   or shorter, but it tries. If usecs is less than USLEEP_BUSYWAIT_USECS,
   busy-wait; otherwise, use nanosleep(). */
void l_usleep(uint64 usecs)
{
    uint64 target;
    struct timespec ts;
    int result;

    if (usecs < USLEEP_BUSYWAIT_USECS) {
        target = usecs_wall() + usecs;
        while (usecs_wall() < target)
            /* nothing */ ;
    } else {
        ts.tv_sec = usecs / ONE_MILLION;
        ts.tv_nsec = (usecs % ONE_MILLION) * 1000;
        result = nanosleep(&ts, NULL);
        l_ASSERT(result == 0);
    }
}

/* Initialize some globals. */
void timing_init(freq sr, unsigned tl, bool limit)
{
    sample_rate = sr;
    tick_len = tl;
    limit_realtime = limit;
}

/* Called on pause->go transition; set variables appropriately. */
void timing_pausego(void)
{
    go_samples = 0;
    go_usecs = 0;
    go_entry_wall = usecs_wall();
}

/* Called on stop->go transition; set variables appropriately. */
void timing_stopgo(void)
{
    samples_elapsed = 0;

    timing_pausego();
}

/* Called on start of a tick; set up some timing stuff. */
void timing_tick_start(void)
{
    tick_start_wall = usecs_wall();
}

/* Called on end of a tick. More timing stuff. If realtime limiting is on and
   we're ahead of schedule, wait until we were supposed to finish this tick
   (leaving some slush time). */
void timing_tick_end(void)
{
    uint64 allowed, actual;

    go_usecs += usecs_wall() - tick_start_wall;
    go_samples += samples_per_tick();
    samples_elapsed += samples_per_tick();

    if (limit_realtime) {
        allowed = go_samples * seconds_per_sample() * ONE_MILLION;
        actual = usecs_wall() - go_entry_wall + SLUSH_USECS;

        if (allowed > actual)
            l_usleep(allowed - actual);
    }
}

/* Print some statistics about time. */
void timing_stats_pr(void)
{
    pr("Since entry of go state from stop:\n");
    pr("  Samples calculated:        %llu\n", samples_elapsed);
    pr("Since entry of go state:\n");
    pr("  Samples calculated:        %llu\n", go_samples);
    pr("  Seconds spent calculating: %5.3f\n",
       (double) go_usecs / ONE_MILLION);
    pr("  Total seconds elapsed:     %5.3f\n",
       (double) (usecs_wall() - go_entry_wall) / ONE_MILLION);
    pr("  Calculation load:          %4.2f\n",
       (double) go_usecs / (go_samples * seconds_per_sample() * ONE_MILLION));
}

inline freq samples_per_second(void)
{
    return sample_rate;
}

inline dur seconds_per_sample(void)
{
    return (dur) 1.0 / sample_rate;
}

inline unsigned samples_per_tick(void)
{
    return tick_len;
}

inline unsigned samples_calc(void)
{
    return samples_elapsed;
}

inline dur seconds_calc(void)
{
    return (dur) samples_elapsed / sample_rate;
}

