/* $Id: types.c,v 1.6 2001/07/20 21:42:03 reid Exp $ */

/* This file implements the type management functions. */

#include "lambda.h"
#include "lcore.h"

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/*** Types ***/

/* These variables store the sizes and human-readable names of each of the
   base types and can be indexed by the names of the raw types; e.g.,
   _l_sizeof[LTYP_INT] == sizeof(int). This depends on the fact that arrays
   start at zero and by default enums start at zero as well and increment by
   one for each item. Hence ORDER IS VERY IMPORTANT and must match exactly
   that of the _ltyp enum. */
size_t _l_sizeof[] = {
    sizeof(bool),
    sizeof(byte),
    sizeof(int),
    sizeof(long),
    sizeof(int16),
    sizeof(uint16),
    sizeof(int32),
    sizeof(uint32),
    sizeof(int64),
    sizeof(uint64),
    sizeof(float),
    sizeof(double),
    sizeof(sample),
    sizeof(freq),
    sizeof(dur),
    sizeof(level)
};
char * _ltyp_str[] = {
    "bool",
    "byte",
    "int",
    "long",
    "int16",
    "uint16",
    "int32",
    "uint32",
    "int64",
    "uint64",
    "float",
    "double",
    "sample",
    "freq",
    "dur",
    "level"
};

/*** Public functions ***/

/* Return the number of bytes required to store an object described by t. */
size_t l_sizeof(ltyp * t)
{
    size_t sz;

    sz = _l_sizeof[t->raw_type];

    switch (t->meta_type) {
    case SCALAR:
        return sz;
        break;
    case ARRAY:
        return sz * t->n_elements;
        break;
    case ARRAY_REL:
        return sz * t->n_elements * samples_per_tick();
        break;
    }
    
    l_ASSERT(0);
    return 0;
}

/* Allocate and initialize a new ltyp object, and return a pointer to it.
   Cannot fail. */
ltyp * ltyp_new(meta_ltyp meta_type, raw_ltyp raw_type, unsigned n_elements)
{
    ltyp * t;
    
    l_ASSERT(meta_type >= SCALAR && meta_type <= ARRAY_REL);
    l_ASSERT(!(meta_type == ARRAY && n_elements == 0));
    l_ASSERT(!(meta_type == ARRAY_REL && n_elements == 0));
    l_ASSERT(raw_type >= LTYP_INT && raw_type <= LTYP_LEVEL);

    t = l_malloc(sizeof(ltyp));
    t->meta_type = meta_type;
    t->raw_type = raw_type;
    t->n_elements = n_elements;

    return t;
}

/* Free t; postcondition: t points to garbage. */
void ltyp_free(ltyp * t)
{
    l_free(t);
}

/* Return 1 if t1 and t2 are equivalent, 0 otherwise. */
bool ltyp_eq(ltyp * t1, ltyp * t2)
{
    if (   t1->meta_type == t2->meta_type
        && t1->raw_type == t2->raw_type
        && t1->n_elements == t2->n_elements)
        return 1;
    else
        return 0;
}

/* Create a human-readable string representation of t (e.g. "int"), and store
   it in freshly allocated memory (it is the caller's responsibility to free
   the memory). Return a pointer to it. Cannot fail. */
char * ltyp_str(ltyp * t)
{
    char * str, * subscript;

    if (t->meta_type == ARRAY || t->meta_type == ARRAY_REL) {
        /* check out the fab ternary operator */
        asprintf(&subscript, "[%u%s]", t->n_elements,
                 (t->meta_type == ARRAY_REL) ? "r" : "");
        if (subscript == NULL) {
            error_pr("asprintf() failed. This means FIXME.\n");
            abort();
        }
    } else {
        subscript = "";
    }

    asprintf(&str, "%s%s", _ltyp_str[t->raw_type], subscript);
    if (str == NULL) {
        error_pr("asprintf() failed. This means FIXME.\n");
        abort();
    }
    if (strcmp(subscript, ""))
        l_free(subscript);

    return str;
}

/* Create an appropriate string representation of *value according to t, and
   store it in freshly allocated memory (it is the caller's responsibility to
   free the memory). Return a pointer to it. Cannot fail.

   FIXME [bug 36]: sample, freq, dur, and level should be printed more
   elegantly. */
char * ltyp_tostr(ltyp * t, void * value)
{
    char * str;

    switch (t->raw_type) {
    case LTYP_BOOL:
        asprintf(&str, "%d", *((bool *) value));
        break;
    case LTYP_BYTE:
        asprintf(&str, "%u", *((byte *) value));
        break;
    case LTYP_INT:
        asprintf(&str, "%d", *((int *) value));
        break;
    case LTYP_LONG:
        asprintf(&str, "%ld", *((long *) value));
        break;
    case LTYP_INT16:
        asprintf(&str, "%d", *((int16 *) value));
        break;
    case LTYP_UINT16:
        asprintf(&str, "%u", *((uint16 *) value));
        break;
    case LTYP_INT32:
        asprintf(&str, "%ld", (long) *((int32 *) value));
        break;
    case LTYP_UINT32:
        asprintf(&str, "%lu", (unsigned long) *((uint32 *) value));
        break;
    case LTYP_INT64:
    case LTYP_UINT64:
        /* FIXME [bug 36] */
        l_ASSERT(0);
        break;
    case LTYP_FLOAT:
        asprintf(&str, "%g", *((float *) value));
        break;
    case LTYP_DOUBLE:
        asprintf(&str, "%g", *((double *) value));
        break;
    case LTYP_SAMPLE:
        asprintf(&str, "%g", *((sample *) value));
        break;
    case LTYP_FREQ:
        asprintf(&str, "%g", *((freq *) value));
        break;
    case LTYP_DUR:
        asprintf(&str, "%g", *((dur *) value));
        break;
    case LTYP_LEVEL:
        asprintf(&str, "%g", *((level *) value));
        break;
    default:
        l_ASSERT(0);
    }

    if (str == NULL) {
        error_pr("asprintf() failed. This means FIXME.\n");
        abort();
    }
    
    return str;
}

/* Interpret str appropriately according to t; store the result in buf. If buf
   is too small to hold a value of the type described by t, the results are
   undefined. On success, return OK; otherwise, set the error string, do not
   modify the contents of buf, and return ERR.

   FIXME [bug 36]: self-allocate memory, since ltyp_tostr() does it.
   FIXME [bug 36]: interpret arrays too.
   FIXME [bug 36]: do range checking. */
lerr ltyp_fromstr(void * buf, ltyp * t, char * str)
{
    byte val[L_SIZEOF_MAX];
    char * tailptr;

    if (t->meta_type != SCALAR) {
        err_set("ltyp_fromstr() cannot translate arrays");
        return ERR;
    }

    errno = 0;
    switch (t->raw_type) {
    case LTYP_BOOL:
        *((bool *) val) = (bool) strtol(str, &tailptr, 0);
        if (*((bool *) val) != 0 && *((bool *) val) != 1) {
            err_set("boolean value must be 1 or 0");
            return ERR;
        }
        break;
    case LTYP_BYTE:
        *((byte *) val) = (byte) strtoul(str, &tailptr, 0);
        break;
    case LTYP_INT:
        *((int *) val) = (int) strtol(str, &tailptr, 0);
        break;
    case LTYP_LONG:
        *((long *) val) = (long) strtol(str, &tailptr, 0);
        break;
    case LTYP_INT16:
        *((int16 *) val) = (int16) strtol(str, &tailptr, 0);
        break;
    case LTYP_UINT16:
        *((uint16 *) val) = (uint16) strtoul(str, &tailptr, 0);
        break;
    case LTYP_INT32:
        *((int32 *) val) = (int32) strtol(str, &tailptr, 0);
        break;
    case LTYP_UINT32:
        *((uint32 *) val) = (uint32) strtoul(str, &tailptr, 0);
        break;
    case LTYP_INT64:
    case LTYP_UINT64:
        /* FIXME [bug 36] */
        l_ASSERT(0);
        break;
    case LTYP_FLOAT:
        *((float *) val) = (float) strtod(str, &tailptr);
        break;
    case LTYP_DOUBLE:
        *((double *) val) = (double) strtod(str, &tailptr);
        break;
    case LTYP_SAMPLE:
        *((sample *) val) = (sample) strtod(str, &tailptr);
        break;
    case LTYP_FREQ:
        *((freq *) val) = (freq) strtod(str, &tailptr);
        break;
    case LTYP_DUR:
        *((dur *) val) = (dur) strtod(str, &tailptr);
        break;
    case LTYP_LEVEL:
        *((level *) val) = (level) strtod(str, &tailptr);
        break;
    default:
        l_ASSERT(0);
    }

    if (errno != 0) {
        err_set("%s", strerror(errno));
        return ERR;
    }

    if (*tailptr != '\0') {
        err_set("junk at end of string");
        return ERR;
    }

    memcpy(buf, val, l_sizeof(t));
    return OK;
}
