/* $Id: io-oss.c,v 1.6 2001/04/20 05:48:27 reid Exp $ */

/* Reference: Open Sound System Programmer's Guide, v1.11, 2000/11/07
              http://www.opensound.com/pguide/oss.pdf */

/* FIXME:

   - Set better fragment size.
   - Implement proper full duplex.
   - Other real time stuff.
   - Optimize for speed.
   - Provide for device file selection at runtime. */

#include "lambda.h"

#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>


/*** Constants ***/

/* which device file to use? */
#define OSS_DEVICE "/dev/dsp"


/*** Identification ***/

unsigned magic = MOD_MAGIC_1_0_0;
char name[] = "io-oss";
char version[] = "1.0d1";
char description[] = "Plays and records sound via the OSS drivers.";


/*** Ports ***/

port_decl port_decls[] = {
    { "play-right",   IN, ARRAY_REL, LTYP_SAMPLE, 1, NONSTRICT,
                      { } },
    { "play-left",    IN, ARRAY_REL, LTYP_SAMPLE, 1, NONSTRICT,
                      { "play-right" } },
    { "record-right", OUT,  ARRAY_REL, LTYP_SAMPLE, 1, NONSTRICT, 
                      { } },
    { "record-left",  OUT,  ARRAY_REL, LTYP_SAMPLE, 1, NONSTRICT,
                      { "record-right" } },
    END_PORT_DECLARATIONS
};


/*** Types ***/

typedef struct {
    port * play_r, * play_l;
    port * rec_r, * rec_l;
    int fd;
    bool play_active, play_stereo, rec_active, rec_stereo;
    int16 * buf_oss;
    sample * buf_r, * buf_l;
} context_t;


/*** Private prototypes ***/

lerr config_dsp(context_t * x);


/*** Public functions ***/

lerr lm_go(void ** context, unsigned n_ports, port * ports[])
{
    context_t * x;

    x = l_malloc(sizeof(context_t));
    x->play_active = 0;
    x->play_stereo = 0;
    x->rec_active = 0;
    x->rec_stereo = 0;

    x->play_r = port_addrof("play-right", n_ports, ports);
    l_ASSERT(x->play_r != NULL && x->play_r->direction == IN);
    if (port_ready(x->play_r))
        x->play_active = 1;
    x->play_l = port_addrof("play-left", n_ports, ports);
    l_ASSERT(x->play_l != NULL && x->play_l->direction == IN);
    if (port_ready(x->play_l))
        x->play_stereo = 1;

    x->rec_r = port_addrof("record-right", n_ports, ports);
    l_ASSERT(x->rec_r != NULL && x->rec_r->direction == OUT);
    if (port_ready(x->rec_r))
        x->rec_active = 1;
    x->rec_l = port_addrof("record-left", n_ports, ports);
    l_ASSERT(x->rec_l != NULL && x->rec_r->direction == OUT);
    if (port_ready(x->rec_l))
        x->rec_stereo = 1;

    l_ASSERT(x->play_active || x->rec_active);

    if (config_dsp(x) != OK) {
        /* config_dsp will have already set the error string */
        l_free(x);
        return ERR;
    }

    x->buf_r = l_malloc(sizeof(sample) * samples_per_tick());
    x->buf_oss = l_malloc(sizeof(int16) * samples_per_tick());
    if (x->play_stereo || x->rec_stereo) {
        x->buf_l = l_malloc(sizeof(sample) * samples_per_tick());
        x->buf_oss = l_realloc(x->buf_oss,
                               2 * sizeof(int16) * samples_per_tick());
    }

    *context = x;
    
    return OK;
}

lerr lm_pause(void * context)
{
    context_t * x = context;

    if (ioctl(x->fd, SNDCTL_DSP_POST) == -1) {
        err_set("pause failed for %s: %s", OSS_DEVICE, strerror(errno));
        return ERR;
    }
    
    return OK;
}

lerr lm_stop(void * context)
{
    context_t * x = context;

    close(x->fd);
    l_free(x->buf_oss);
    l_free(x->buf_r);
    if (x->play_stereo || x->rec_stereo)
        l_free(x->buf_l);
    l_free(x);

    return OK;
}

lerr lm_tick(void * context)
{
    context_t * x = context;
    int i, offset;

    /* In the loops below, conversions from sample to int16 are implicit. */

    if (x->play_active) {
        l_recv(x->play_r, x->buf_r);
        if (x->play_stereo)
            l_recv(x->play_l, x->buf_l);
        
        offset = 0;
        for (i = 0; i < samples_per_tick(); i++) {
            /* left sample comes first */
            if (x->play_stereo) {
                x->buf_oss[offset] = x->buf_l[i] * SHRT_MAX;
                offset++;
            }

            x->buf_oss[offset] = x->buf_r[i] * SHRT_MAX;
            offset++;
        }

        /* FIXME: handle errors */
        write(x->fd, x->buf_oss,
              sizeof(int16) * samples_per_tick() * (x->play_stereo ? 2 : 1));
    }

    /* FIXME: recording not implemented! */

    return OK;
}


/*** Private functions ***/

/* Configure the sound card. */
lerr config_dsp(context_t * x)
{
    int i, flags, width, channels, rate;

    /* parameters */
    width = AFMT_S16_NE; /* 16 bits signed, host endian */
    channels = x->play_stereo ? 2 : 1;
    rate = samples_per_second();
    debug_pr("io-oss: configuring %s for 16 bits, %d channels, %d Hz\n",
             OSS_DEVICE, channels, rate);


    /* open the audio device */
    if (x->play_active && x->rec_active)
        flags = O_RDWR;
    else if (x->play_active)
        flags = O_WRONLY;
    else
        flags = O_RDONLY;
    x->fd = open(OSS_DEVICE, flags);
    if (x->fd == -1) {
        err_set("can't open %s: %s", OSS_DEVICE, strerror(errno));
        return ERR;
    }
    /* set sample width to 16 bits, native endian */
    i = width;
    if (ioctl(x->fd, SNDCTL_DSP_SETFMT, &i) == -1) {
        err_set("can't set %s to 16 bits: %s", OSS_DEVICE, strerror(errno));
        return ERR;
    }
    if (i != width) {
        err_set("%s does not support 16 bits", OSS_DEVICE);
        return ERR;
    }
    /* set mono/stereo as appropriate */
    i = channels;
    if (ioctl(x->fd, SNDCTL_DSP_CHANNELS, &i) == -1) {
        err_set("can't set %s to %d channels: %s", OSS_DEVICE, channels, 
                strerror(errno));
        return ERR;
    }
    if (i != channels) {
        err_set("%s does not support %d channels", OSS_DEVICE, channels);
        return ERR;
    }
    /* set sample rate */
    i = rate;
    if (ioctl(x->fd, SNDCTL_DSP_SPEED, &i) == -1) {
        err_set("can't set %s to %d Hz: %s", OSS_DEVICE, rate,
                strerror(errno));
        return ERR;
    }
    if (i != rate) {
        err_set("%s does not support %d Hz", OSS_DEVICE, rate);
        return ERR;
    }

    /* whew */
    return OK;
}
