/* $Id: synth-additive.c,v 1.3 2001/04/22 18:59:44 reid Exp $ */

#include "lambda.h"


/*** Constants ***/

/* If you change this you must change port_decls also. */
#define HARMONICS 16


/*** Identification ***/

unsigned magic = MOD_MAGIC_1_0_0;
char name[] = "synth-additive";
char version[] = "1.0d1";
char description[] = "Simple additive synthesizer with 16 partials.";


/*** Ports ***/

port_decl port_decls[] = {
    { "out",          OUT, ARRAY_REL, LTYP_SAMPLE, 1, STRICT, {} },
    { "out-volume",   IN,  SCALAR,    LTYP_LEVEL,  0, STRICT, {} },
    { "f",            IN,  SCALAR,    LTYP_FREQ,   0, STRICT, {} },
    { "fund-volume",  IN,  SCALAR,    LTYP_LEVEL,  0, STRICT, {} },
    { "par2-volume",  IN,  SCALAR,    LTYP_LEVEL,  0, NONSTRICT,
      { "fund-volume" } },
    { "par3-volume",  IN,  SCALAR,    LTYP_LEVEL,  0, NONSTRICT,
      { "par2-volume" } },
    { "par4-volume",  IN,  SCALAR,    LTYP_LEVEL,  0, NONSTRICT,
      { "par3-volume" } },
    { "par5-volume",  IN,  SCALAR,    LTYP_LEVEL,  0, NONSTRICT,
      { "par4-volume" } },
    { "par6-volume",  IN,  SCALAR,    LTYP_LEVEL,  0, NONSTRICT,
      { "par5-volume" } },
    { "par7-volume",  IN,  SCALAR,    LTYP_LEVEL,  0, NONSTRICT,
      { "par6-volume" } },
    { "par8-volume",  IN,  SCALAR,    LTYP_LEVEL,  0, NONSTRICT,
      { "par7-volume" } },
    { "par9-volume",  IN,  SCALAR,    LTYP_LEVEL,  0, NONSTRICT,
      { "par8-volume" } },
    { "par10-volume", IN,  SCALAR,    LTYP_LEVEL,  0, NONSTRICT,
      { "par9-volume" } },
    { "par11-volume", IN,  SCALAR,    LTYP_LEVEL,  0, NONSTRICT,
      { "par10-volume" } },
    { "par12-volume", IN,  SCALAR,    LTYP_LEVEL,  0, NONSTRICT,
      { "par11-volume" } },
    { "par13-volume", IN,  SCALAR,    LTYP_LEVEL,  0, NONSTRICT,
      { "par12-volume" } },
    { "par14-volume", IN,  SCALAR,    LTYP_LEVEL,  0, NONSTRICT,
      { "par13-volume" } },
    { "par15-volume", IN,  SCALAR,    LTYP_LEVEL,  0, NONSTRICT,
      { "par14-volume" } },
    { "par16-volume", IN,  SCALAR,    LTYP_LEVEL,  0, NONSTRICT,
      { "par15-volume" } },
    END_PORT_DECLARATIONS
};


/*** Types ***/

typedef struct {
    port * out;
    port * out_vol;
    port * f;
    port * har_vols[HARMONICS];
    osc * oscs[HARMONICS];
    unsigned harmonics;
    sample * buf;
    sample * buf_a;
} context_t;


/*** Public functions ***/

lerr lm_go(void ** context, unsigned n_ports, port * ports[])
{
    context_t * x;
    char * s;
    int i;

    x = l_malloc(sizeof(context_t));
    x->out = port_addrof("out", n_ports, ports);
    x->out_vol = port_addrof("out-volume", n_ports, ports);
    x->f = port_addrof("f", n_ports, ports);

    x->har_vols[0] = port_addrof("fund-volume", n_ports, ports);
    for (i = 1; i < HARMONICS; i++) {
        /* FIXME: check s for NULL. */
        asprintf(&s, "par%d-volume", i + 1);
        x->har_vols[i] = port_addrof(s, n_ports, ports);
        l_free(s);
        if (!port_ready(x->har_vols[i]))
            break;
    }
    x->harmonics = i;

    for (i = 0; i < x->harmonics; i++)
        x->oscs[i] = osc_new(0.0, 0.0);

    x->buf = l_malloc(sizeof(sample) * samples_per_tick());
    x->buf_a = l_malloc(sizeof(sample) * samples_per_tick());

    *context = x;
    
    return OK;
}

lerr lm_pause(void * context)
{
    return OK;
}

lerr lm_stop(void * context)
{
    context_t * x = context;
    int i;

    l_free(x->buf);
    l_free(x->buf_a);
    for (i = 0; i < x->harmonics; i++)
        osc_free(x->oscs[i]);
    l_free(x);

    return OK;
}

lerr lm_tick(void * context)
{
    context_t * x = context;
    freq f;
    level out_vol;
    level har_vols[HARMONICS];
    int i, j;

    l_recv(x->f, &f);
    l_recv(x->out_vol, &out_vol);
    for (i = 0; i < x->harmonics; i++)
        l_recv(x->har_vols[i], &(har_vols[i]));

    for (j = 0; j < samples_per_tick(); j++)
        x->buf_a[j] = 0.0;
    for (i = 0; i < x->harmonics; i++) {
        /* Bail if we go above the Nyquist frequency. 
           FIXME: perhaps warn somehow about this? */
        if (f * (i + 1) > samples_per_second() / 2)
            break;
        osc_sine(x->buf, x->oscs[i], f * (i + 1), samples_per_tick());
        for (j = 0; j < samples_per_tick(); j++)
            x->buf_a[j] += x->buf[j] * har_vols[i];
    }

    for (j = 0; j < samples_per_tick(); j++)
        x->buf_a[j] *= out_vol;
    l_send(x->out, x->buf_a);

    return OK;
}
